class ocrUtils {
  // 构造器方法，用于初始化新创建的对象
  constructor(name) {
    this.name = name;
  }
  /**
   * @type {ocrController}
   * @static
   */
  static ocrController
  /** @type {String} */
  static mainPath
  // 0都没有
// 1都有
// 2image
// 3text
  static init(){
    this.app = Application.sharedInstance()
    this.data = Database.sharedInstance()
    this.currentWindow = this.app.focusWindow
    this.version = this.appVersion()
  }
  static get notebookController(){
    return this.studyController.notebookController
  }
  static get studyController(){
    return this.app.studyController(this.currentWindow)
  }
  static get studyView() {
    return this.app.studyController(this.currentWindow).view
  }
  static get currentNotebookId() {
    return this.studyController.notebookController.notebookId
  }
  static showHUD(message,duration=2) {
    this.app.showHUD(message,this.currentWindow,2)
  }

  static appVersion() {
    let info = {}
    let version = parseFloat(this,this.app.appVersion)
    if (version >= 4) {
      info.version = "marginnote4"
    }else{
      info.version = "marginnote3"
    }
    switch (this.app.osType) {
      case 0:
        info.type = "iPadOS"
        break;
      case 1:
        info.type = "iPhoneOS"
        break;
      case 2:
        info.type = "macOS"
      default:
        break;
    }
    return info
  }
  static getByDefault(key,defaultValue) {
    let value = NSUserDefaults.standardUserDefaults().objectForKey(key)
    if (value === undefined) {
      NSUserDefaults.standardUserDefaults().setObjectForKey(defaultValue,key)
      return defaultValue
    }
    return value
  }
  static  getNoteColors() {
    return ["#ffffb4","#ccfdc4","#b4d1fb","#f3aebe","#ffff54","#75fb4c","#55bbf9","#ea3323","#ef8733","#377e47","#173dac","#be3223","#ffffff","#dadada","#b4b4b4","#bd9fdc"]
  }
  static getNoteById(noteid) {
    let note = this.data.getNoteById(noteid)
    return note
  }
  static getNoteBookById(notebookId) {
    let notebook = this.data.getNotebookById(notebookId)
    return notebook
  }
  static getUrlByNoteId(noteid) {
    let ver = this.appVersion()
    return ver.version+'app://note/'+noteid
  }
  /**
   * 
   * @param {String} url 
   * @returns {String}
   */
  static getNoteIdByURL(url) {
    let targetNoteId = url.trim()
    if (/^marginnote\dapp:\/\/note\//.test(targetNoteId)) {
      targetNoteId = targetNoteId.slice(22)
    }
    return targetNoteId
  }
  static clipboardText() {
    return UIPasteboard.generalPasteboard().string
  }
  static copy(text) {
    UIPasteboard.generalPasteboard().string = text
  }
  static copyJSON(object) {
    UIPasteboard.generalPasteboard().string = JSON.stringify(object,null,2)
  }
  /**
   * 
   * @param {NSData} imageData 
   */
  static copyImage(imageData) {
    UIPasteboard.generalPasteboard().setDataForPasteboardType(imageData,"public.png")
  }

  
  static get currentDocController() {
    return this.studyController.readerController.currentDocumentController
  }
  static currentNotebook() {
    let notebookId = this.studyController.notebookController.notebookId
    return this.getNoteBookById(notebookId)
  }
  static undoGrouping(f,notebookId = this.currentNotebookId){
    UndoManager.sharedInstance().undoGrouping(
      String(Date.now()),
      notebookId,
      f
    )
    this.app.refreshAfterDBChanged(notebookId)
  }
  static checkOCRController(){
    let studyView = this.studyView
    if (!this.ocrController) {
      this.ocrController = ocrController.new();
    }
    if (!this.ocrController.view.isDescendantOfView(studyView)) {
      studyView.addSubview(this.ocrController.view)
    }
  }
  static getImage(path,scale=2) {
    return UIImage.imageWithDataScale(NSData.dataWithContentsOfFile(path), scale)
  }
  static refreshAddonCommands(){
    this.studyController.refreshAddonCommands()
  }
  static addObserver(observer,selector,name){
    NSNotificationCenter.defaultCenter().addObserverSelectorName(observer, selector, name);
  }
  static removeObserver(observer,name){
    NSNotificationCenter.defaultCenter().removeObserverName(observer, name);
  }
  static checkSender(sender,window){
    return this.app.checkNotifySenderInWindow(sender, window)
  }
  static getPopoverAndPresent(sender,commandTable,width=100) {
    var menuController = MenuController.new();
    menuController.commandTable = commandTable
    menuController.rowHeight = 35;
    menuController.preferredContentSize = {
      width: width,
      height: menuController.rowHeight * menuController.commandTable.length
    };
    //右 0
    //下 1，3
    //上 2
    var popoverController = new UIPopoverController(menuController);
    var r = sender.convertRectToView(sender.bounds,this.studyView);
    popoverController.presentPopoverFromRect(r, this.studyView, 1, true);
    return popoverController
  }
  static hexColorAlpha(hex,alpha) {
    let color = UIColor.colorWithHexString(hex)
    return alpha!==undefined?color.colorWithAlphaComponent(alpha):color
  }
  static getMediaByHash(hash) {
    let image = this.data.getMediaByHash(hash)
    return image
  }
  static 
  getFocusNote() {
    let notebookController = this.notebookController
    // copy("text:"+!notebookController.view.hidden)
    if (!notebookController.view.hidden && notebookController.mindmapView && notebookController.focusNote) {
      return notebookController.focusNote
    }else{
      return (this.currentDocController.focusNote ?? this.currentDocController.lastFocusNote)
    }
  }
  /**
   * 
   * @param {MbBookNote} note 
   * @returns 
   */
  static getImageFromNote(note) {
    if (note.excerptPic) {
      return ocrUtils.getMediaByHash(note.excerptPic.paint)
    }
    if (note.comments.length) {
      let imageData = undefined
      for (let i = 0; i < note.comments.length; i++) {
        const comment = note.comments[i];
        if (comment.type === 'PaintNote' && comment.paint) {
          imageData = ocrUtils.getMediaByHash(comment.paint)
          break
        }
        
      }
      if (imageData) {
        return imageData
      }
    }
    return undefined
  }
  static setFrame(target,frame){
    target.view.frame = frame
    target.currentFrame = frame
  }
  static getRandomElement(arr) {
    if (arr && arr.length === 1) {
      return arr[0]
    }
    if (arr && arr.length) {
      const randomIndex = Math.floor(Math.random() * arr.length);
      return arr[randomIndex];
    }
    return ""; // 或者抛出一个错误，如果数组为空或者未定义
  }
  static getImageForOCR(){
    let foucsNote = ocrUtils.getFocusNote()
    let imageData = ocrUtils.currentDocController.imageFromSelection()
    if (!imageData && foucsNote) {
      imageData = ocrUtils.getImageFromNote(foucsNote)
    }
    if (!imageData) {
      imageData = ocrUtils.currentDocController.imageFromFocusNote()
    }
    return imageData
  }
  /**
   * 
   * @param {string} fullPath 
   * @returns {string}
   */
  static getExtensionFolder(fullPath) {
      // 找到最后一个'/'的位置
      let lastSlashIndex = fullPath.lastIndexOf('/');
      // 从最后一个'/'之后截取字符串，得到文件名
      let fileName = fullPath.substring(0,lastSlashIndex);
      return fileName;
  }
  static checkMNUtilsFolder(fullPath){
    let extensionFolder = this.getExtensionFolder(fullPath)
    return NSFileManager.defaultManager().fileExistsAtPath(extensionFolder+"/marginnote.extension.mnutils/main.js")
  }
}

class ocrNetwork {
  constructor(name) {
    this.name = name;
  }
  static genNSURL(url) {
    return NSURL.URLWithString(url)
  }
  static initRequest(url,options) {
    const request = NSMutableURLRequest.requestWithURL(this.genNSURL(url))
    request.setHTTPMethod(options.method ?? "GET")
    request.setTimeoutInterval(options.timeout ?? 10)
    const headers = {
      "User-Agent":
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.1 Safari/605.1.15",
      "Content-Type": "application/json",
      Accept: "application/json"
    }
    request.setAllHTTPHeaderFields({
      ...headers,
      ...(options.headers ?? {})
    })
    if (options.search) {
      request.setURL(
        this.genNSURL(
          `${url.trim()}?${Object.entries(options.search).reduce((acc, cur) => {
            const [key, value] = cur
            return `${acc ? acc + "&" : ""}${key}=${encodeURIComponent(value)}`
          }, "")}`
        )
      )
    } else if (options.body) {
      request.setHTTPBody(NSData.dataWithStringEncoding(options.body, 4))
    } else if (options.form) {
      request.setHTTPBody(
        NSData.dataWithStringEncoding(
          Object.entries(options.form).reduce((acc, cur) => {
            const [key, value] = cur
            return `${acc ? acc + "&" : ""}${key}=${encodeURIComponent(value)}`
          }, ""),
          4
        )
      )
    } else if (options.json) {
      request.setHTTPBody(
        NSJSONSerialization.dataWithJSONObjectOptions(
          options.json,
          1
        )
      )
    }
    return request
  }
  static async sendRequest(request,msg="OCR",showMsg=true) {
    const queue = NSOperationQueue.mainQueue()
    // MNUtil.showHUD("send request")
    return new Promise((resolve, reject) => {
      NSURLConnection.sendAsynchronousRequestQueueCompletionHandler(
        request,
        queue,
        (res, data, err) => {
          // let response = {}
          // MNUtil.copyJSON("result:"+res.statusCode())
          if (res.statusCode() === 200) {
            if (showMsg) {
              ocrUtils.showHUD(msg+" success")
            }
          }else{
            MNUtil.copy("Error in "+msg+":; statusCode: "+res.statusCode())
            ocrUtils.showHUD("Error in "+msg+":; statusCode: "+res.statusCode() )
          }
          const result = NSJSONSerialization.JSONObjectWithDataOptions(
            data,
            1<<0
          )
            MNUtil.copyJSON(result)

          if (NSJSONSerialization.isValidJSONObject(result)){
            if (err.localizedDescription){
              // response.error = err.localizedDescription
              // ocrUtils.copyJSON(res)
              result.success = false
              ocrUtils.showHUD(err.localizedDescription)
              // ocrUtils.copy(err.localizedDescription)
              // resolve({error:err.localizedDescription})
            }
            resolve(result)
          }
          resolve(result)
        }
      )
  })
  }
  static async fetch (url,options = {},msg,showMsg){
    const request = this.initRequest(url, options)
    const res = await this.sendRequest(request,msg,showMsg)
    return res
  }
/**
 * 
 * @returns {Promise<Object>}
 */
 static async simpleTexOCR(imageData) {
  ocrUtils.showHUD("OCR By simpleTex")
  let key = ocrConfig.getConfig("simpleTexApikey").trim()
  if (!key) {
    if (this.checkSubscribe()) {
      let res = await subscriptionNetwork.getKey("simpletex")
      // MNUtil.copyJSON(res)
      if (!res) {
        return undefined
      }
      key = ocrUtils.getRandomElement(res.keys)
      // ocrUtils.copy(key)
      // return undefined
    }else{
      return undefined
    }
  }
  if (!key) {
    ocrUtils.showHUD("No simpleTex API key")
    return
  }
  const headers = {
    token: key
  }
  let url = "https://server.simpletex.cn/api/simpletex_ocr"
  // if (ocrConfig.getConfig("simpleTexGeneral")) {
  //   url = "https://server.simpletex.cn/api/simpletex_ocr"
  // }else{
  //   url = ocrConfig.getConfig("simpleTexTurbo") ?"https://server.simpletex.cn/api/latex_ocr_turbo" :"https://server.simpletex.cn/api/latex_ocr"
  // }
  const request = this.initOCRRequest(url, {
      headers: headers
    },
    imageData)
  try {
    let res = await this.sendRequest(request)
    ocrUtils.showHUD("OCR success")
    let ocrResult
    if (res.res.type === "formula") {
      ocrResult = "$$"+res.res.info.trim()+"$$"
    }
    if (res.res.type === "doc") {
      ocrResult = res.res.info.markdown.trim()
    }
    return ocrResult
    
  } catch (error) {
    ocrUtils.showHUD(error)
    return undefined
  }
}
  /**
   * 
   * @returns {Boolean}
   */
  static checkSubscribe(){
    // return true
    if (typeof subscriptionConfig !== 'undefined') {
      let res = subscriptionConfig.checkSubscribed(false,true)
      // MNUtil.copy(res)
      return res
    }else{
      this.showHUD("Set your API key or install 'MN Subscrption'")
      return false
    }
  }
  static async refreshAccessToken(refreshToken){
    MNUtil.showHUD("get access_token")
    let url = "https://api.doc2x.noedgeai.com/api/token/refresh"
    let headers = {
      Authorization: "Bearer "+refreshToken
    }
    let res = await this.fetch(url,{
      headers:headers,
      method: "Post"
      },"refresh access_token")
    if (res.msg === "ok") {
      return res.data.token
    }
    return undefined
  }
  static async getAccessToken(){
    let accessToken
    let key = ocrConfig.getConfig("doc2xApikey").trim()
    if (!key) {
      if (this.checkSubscribe()) {
        let res = await subscriptionNetwork.getKey("doc2x")
        if (!res) {
          return undefined
        }
        accessToken = ocrUtils.getRandomElement(res.keys)
        // ocrUtils.copy(key)
        // return undefined
      }else{
        return undefined
      }
    }else{
      if (!this.accessToken) {
        this.accessToken = await this.refreshAccessToken(key)
      }
      accessToken = this.accessToken
    }
    return accessToken
  }
/**
 * 
 * @returns {NSMutableURLRequest}
 */
 static intDoc2xOCRRequest(imageData,accessToken,stream=true) {
  if (!accessToken) {
    return undefined
  }
  ocrUtils.showHUD("OCR By doc2X")
  let config = ocrConfig.config
  const headers = {
    Authorization: "Bearer "+accessToken
  }
  let url = "https://api.doc2x.noedgeai.com/api/"+(stream?"platform/img":"platform/async/img")
  const request = this.initOCRRequest(url, {
      headers: headers
    },
    imageData,
    config
    )
  return request
}
/**
 * 
 * @returns {NSMutableURLRequest}
 */
 static intDoc2xPDFOCRRequest(pdfData,accessToken,config = {}) {
  const headers = {
    Authorization: "Bearer "+accessToken
  }
  let url = "https://api.doc2x.noedgeai.com/api/platform/pdf"
  const request = this.initOCRRequest(url, {
      headers: headers
    },
    pdfData,
    config
    )
  return request
}

/**
 * 
 * @returns {Promise<Object>}
 */
 static async doc2xOCR(imageData) {
  try {
    let accessToken = await this.getAccessToken()
    let request = this.intDoc2xOCRRequest(imageData,accessToken,false)
    if (!request) {
      MNUtil.showHUD("no request")
      return undefined
    }
      // MNUtil.copy("no request")
    // MNUtil.showHUD("message")
    let res = await this.sendRequest(request,"doc2x")
    if (res.code === "success") {
      let uuid = res.data.uuid
      let status = ""
      let md = undefined
      let finish = false
      // MNUtil.copyJSON(res)
      const headers = {
        Authorization: "Bearer "+accessToken
      }
      await MNUtil.delay(1)
      MNUtil.showHUD("Getting image content...")
      do {
        // let url = "https://api.doc2x.noedgeai.com/api/v1/async/status?uuid="+uuid
        let url = "https://api.doc2x.noedgeai.com/api/platform/async/status?uuid="+uuid
        let result = await this.fetch(url,{headers: headers},"",false)
        status = result.data.status
        MNUtil.showHUD("Progress "+result.data.progress+"%")
        if (status === "success") {
          finish = true
          md = result.data.result.pages[0].md
          if (!md) {
            MNUtil.copyJSON(result.data.result.pages[0])
          }
        }
        await MNUtil.delay(0.5)
      } while (!finish);
      if (!md) {
        MNUtil.showHUD("Empty OCR result!")
        return undefined
      }
      let convertText = md
        .replace(/(\\\[\s?)|(\s?\\\])/g, '$$') // Replace display math mode delimiters
        .replace(/(\\\(\s?)|(\s?\\\))/g, '$') 
      return convertText
    }
    return undefined
    
  } catch (error) {
    MNUtil.copy(error)
    MNUtil.showHUD(error)
    return undefined
  }
}
static parseDoc2XRespnse(data){
  let test = CryptoJS.enc.Base64.parse(data.base64Encoding())
  let textString = CryptoJS.enc.Utf8.stringify(test);
  // MNUtil.copy(textString)
  let res = JSON.parse(textString.split("data:").at(-1))
  return res.data
}
static initOCRRequest (url,options,imageData,config={}) {
  const request = NSMutableURLRequest.requestWithURL(this.genNSURL(url))
  // try {
  request.setHTTPMethod("Post")
  request.httpShouldHandleCookies = false
  // request.setCachePolicy(4)
  // request.setTimeoutInterval(options.timeout ?? 10)
  let boundary = NSUUID.UUID().UUIDString()

  const headers = {
    "User-Agent": "curl/8.4.0",
    "Accept-Encoding":"*",
    "Accept-Language":"*",
    "Connection":"close",
    "Content-Type": "multipart/form-data; boundary="+boundary,
    Accept: "*/*"
  }
  request.setAllHTTPHeaderFields({
    ...headers,
    ...(options.headers ?? {})
  })
  let fileName = config.fileName ?? "image.png"
  
  let body = NSMutableData.new()
  let filePart
  // MNUtil.copyJSON(config)
  switch (ocrConfig.getConfig("source")) {
    case "Doc2X":
      if (config.imageCorrection) {
        filePart = NSData.dataWithStringEncoding(`--${boundary}\r\nContent-Disposition: form-data; name="img_correction"\r\n\r\n`, 4)
        body.appendData(filePart)
        body.appendData(NSData.dataWithStringEncoding(`true\r\n`, 4))
      }
      if (config.pureEquation) {
        filePart = NSData.dataWithStringEncoding(`--${boundary}\r\nContent-Disposition: form-data; name="equation"\r\n\r\n`, 4)
        body.appendData(filePart)
        body.appendData(NSData.dataWithStringEncoding(`true\r\n`, 4))
      }
      filePart = NSData.dataWithStringEncoding(`--${boundary}\r\nContent-Disposition: form-data; name="file"; filename="${fileName}"\r\nContent-Type: image/png\r\n\r\n`, 4)
      break;
    case "Doc2XPDF":
      if (config.PDFOCR) {
        filePart = NSData.dataWithStringEncoding(`--${boundary}\r\nContent-Disposition: form-data; name="ocr"\r\n\r\n`, 4)
        body.appendData(filePart)
        body.appendData(NSData.dataWithStringEncoding(`true\r\n`, 4))
      }
      filePart = NSData.dataWithStringEncoding(`--${boundary}\r\nContent-Disposition: form-data; name="file"; filename="${fileName}"\r\nContent-Type: application/pdf\r\n\r\n`, 4)
      break
    case "SimpleTex":
      if (ocrConfig.config.simpleTexRotation) {
        filePart = NSData.dataWithStringEncoding(`--${boundary}\r\nContent-Disposition: form-data; name="enable_img_rot"\r\n\r\n`, 4)
        body.appendData(filePart)
        body.appendData(NSData.dataWithStringEncoding(`true\r\n`, 4))
      }
      if (ocrConfig.config.simpleTexRecMode) {
        // MNUtil.showHUD(ocrConfig.config.simpleTexRecMode)
        filePart = NSData.dataWithStringEncoding(`--${boundary}\r\nContent-Disposition: form-data; name="rec_mode"\r\n\r\n`, 4)
        body.appendData(filePart)
        body.appendData(NSData.dataWithStringEncoding(`${ocrConfig.config.simpleTexRecMode}\r\n`, 4))
      }
      filePart = NSData.dataWithStringEncoding(`--${boundary}\r\nContent-Disposition: form-data; name="file"; filename="${fileName}"\r\nContent-Type: image/png\r\n\r\n`, 4)
      break
    default:
      break;
  }

  body.appendData(filePart)
  body.appendData(imageData)
  
  let endBoundary = NSData.dataWithStringEncoding(`\r\n--${boundary}--\r\n`, 4)
  body.appendData(endBoundary)
  request.setHTTPBody(body)
  return request
}
static async OCR(imageData){
  switch (ocrConfig.getConfig("source")) {
    case "Doc2X":
      return await this.doc2xOCR(imageData)
    case "SimpleTex":
      return await this.simpleTexOCR(imageData)
    default:
      return undefined
  }

}
}

class ocrConfig {
  constructor(name) {
    this.name = name;
  }
  static defaultConfig = {
    source:"Doc2X",
    simpleTexApikey: "",
    simpleTexTurbo:false,
    simpleTexGeneral:true,
    simpleTexRecMode:"auto",
    simpleTexRotation:false,
    doc2xApikey: "",
    imageCorrection:false,
    pureEquation:false,
    PDFOCR:false,
    subscribedDay: 0,
    apikey: "",
    freeUsage:0,
    freeDay:0,
    subscriptionDaysRemain:0
  }
  static defaultFileIds = {}
  static init(){
    this.config = this.getByDefault("MNOCR", this.defaultConfig)
    this.fileIds = this.getByDefault("MNOCR_fileIds", this.defaultFileIds)
    // this.config.subscribedDay = 0
    // this.config.subscriptionDaysRemain = 0
  }
  static getByDefault(key,defaultValue) {
    let value = NSUserDefaults.standardUserDefaults().objectForKey(key)
    if (value === undefined) {
      NSUserDefaults.standardUserDefaults().setObjectForKey(defaultValue,key)
      return defaultValue
    }
    return value
  }
  static getConfig(key){
    if (this.config[key] !== undefined) {
      return this.config[key]
    }else{
      return this.defaultConfig[key]
    }
  }

  static save() {
    NSUserDefaults.standardUserDefaults().setObjectForKey(this.config,"MNOCR")
  }
  static saveFileId(md5,uuid){
    this.fileIds[md5] = uuid
    NSUserDefaults.standardUserDefaults().setObjectForKey(this.fileIds,"MNOCR_fileIds")
  }
  static remove(){
    NSUserDefaults.standardUserDefaults().removeObjectForKey("MNOCR")
  }
}